import os
import sys
import time
from PyQt5.QtCore import Qt, QRect, QSize, QPoint, QSettings
from PyQt5.QtWidgets import (
    QApplication,
    QWidget,
    QMainWindow,
    QAction,
    QHBoxLayout,
    QVBoxLayout,
    QGridLayout,
    QLabel,
    QScrollArea,
    QPushButton,
    QFileDialog,
    QCheckBox,
    QSlider,
    QLineEdit,
    QRubberBand,
    QMessageBox,
    QDoubleSpinBox,
    QComboBox,
    QSpinBox,
    QSizePolicy,
)
from PyQt5.QtGui import QImage, QPixmap, QKeySequence, QPainter, QBrush, QColor
from semmatch.core import NavOptions, templateMatch
from semmatch.image import npToQImage, qImgToNp, drawCoords


# popup messages
def popup(parent, message):
    messagebox = QMessageBox(parent)
    messagebox.setText(message)
    messagebox.show()


class ImageViewer(QScrollArea):
    def __init__(self):
        super().__init__()
        self.initUI()

    def initUI(self):
        self.zoom = 1
        self.originalImg = QImage()
        self.activeImg = QImage()

        # need QLabel to setPixmap for images
        self.label = QLabel(self)
        self._refresh()
        self.setWidget(self.label)

    def _refresh(self):
        # save slider values to calculate new positions after zoom
        hbar = self.horizontalScrollBar()
        vbar = self.verticalScrollBar()

        try:
            hbarRatio = hbar.value() / hbar.maximum()
        except ZeroDivisionError:
            hbarRatio = 0.5
        try:
            vbarRatio = vbar.value() / vbar.maximum()
        except ZeroDivisionError:
            vbarRatio = 0.5

        # resize
        img = self.activeImg.scaled(
            self.zoom * self.activeImg.size(), aspectRatioMode=Qt.KeepAspectRatio
        )
        self.label.setPixmap(QPixmap(img))
        self.label.resize(img.size())
        self.label.repaint()

        hbar.setValue(int(hbarRatio * hbar.maximum()))
        vbar.setValue(int(vbarRatio * vbar.maximum()))

    def _setActiveImg(self, img):
        self.activeImg = img
        self._refresh()

    def newImg(self, img):
        self.zoom = 1
        self.originalImg = img
        self._setActiveImg(self.originalImg)

    def zoomIn(self):
        self.zoom *= 1.25
        self._refresh()

    def zoomOut(self):
        self.zoom *= 0.8
        self._refresh()


class ImageViewerCrop(ImageViewer):
    class ColorRubberBand(QRubberBand):
        def __init__(self, shape, parent):
            super().__init__(shape, parent)

        def paintEvent(self, event):
            painter = QPainter()
            painter.begin(self)
            painter.fillRect(self.rect(), QBrush(QColor(255, 0, 0, 128)))
            painter.end()

    def __init__(self):
        super().__init__()
        self.image = None
        self.searchedImg = QImage()

    def loadImage(self, image: "ndarray"):
        self.zoom = 1
        self.image = image
        self.originalImg = npToQImage(image)
        self.parentWidget().sidebar._clearPts()

        self.zoom = 1
        hsb = self.horizontalScrollBar()
        vsb = self.verticalScrollBar()
        hsb.setValue((hsb.minimum() + hsb.maximum()) // 2)
        vsb.setValue((vsb.minimum() + vsb.maximum()) // 2)
        self._refresh()

    def mousePressEvent(self, mouseEvent):
        self.shiftPressed = QApplication.keyboardModifiers() == Qt.ShiftModifier
        self.center = mouseEvent.pos()
        self.rband = self.ColorRubberBand(QRubberBand.Rectangle, self)
        self.rband.setGeometry(QRect(self.center, QSize()))
        self.rband.show()

    def mouseMoveEvent(self, mouseEvent):
        # unnormalized QRect can have negative width/height
        crop = QRect(2 * self.center - mouseEvent.pos(), mouseEvent.pos()).normalized()
        if self.shiftPressed:
            largerSide = max(crop.width(), crop.height())
            self.rband.setGeometry(
                self.center.x() - largerSide // 2,
                self.center.y() - largerSide // 2,
                largerSide,
                largerSide,
            )
        else:
            self.rband.setGeometry(crop)
        self.repaint()

    def mouseReleaseEvent(self, mouseEvent):
        self.rband.hide()
        crop = self.rband.geometry()
        if self.originalImg.isNull():  # no image loaded in
            return
        # handle single click initializing default QRect selecting entire image
        if crop.height() < 10 and crop.width() < 10:
            return
        # calculate X and Y position in original image
        X = int((self.horizontalScrollBar().value() + crop.x()) / self.zoom)
        Y = int((self.verticalScrollBar().value() + crop.y()) / self.zoom)
        origScaleCropWidth = int(crop.width() / self.zoom)
        origScaleCropHeight = int(crop.height() / self.zoom)
        # save crop
        cropQImage = self.originalImg.copy(
            QRect(X, Y, origScaleCropWidth, origScaleCropHeight)
        )
        sidebar = self.parentWidget().sidebar
        sidebar.crop_template.newImg(cropQImage)


class Sidebar(QWidget):
    def __init__(self):
        super().__init__()
        self.initUI()

    def initUI(self):
        self.width = 230
        self.setFixedWidth(self.width)
        self.sldPrec = 3
        self.thresholdVal = 0.8

        # widgets
        self.crop_template = ImageViewer()
        self.crop_template.setFixedHeight(200)
        self.cbBlurTemp = QCheckBox("Blur template")
        self.cbBlurImg = QCheckBox("Blur image")
        self.slider = QSlider(Qt.Horizontal)
        self.slider.setMaximum(10 ** self.sldPrec)
        self.slider.valueChanged.connect(self._setThreshDisp)
        self.slider.setFixedHeight(25)
        self.threshDisp = QDoubleSpinBox()
        self.threshDisp.setFixedHeight(40)
        self.threshDisp.setMaximum(1)
        self.threshDisp.setSingleStep(0.01)
        self.threshDisp.setDecimals(self.sldPrec)
        self.threshDisp.valueChanged.connect(self._setThreshSlider)
        buttonSearch = QPushButton("Search")
        buttonSearch.clicked.connect(self._templateSearch)
        buttonSearch.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.MinimumExpanding)
        self.searchLay = QGridLayout()
        self.searchLay.addWidget(self.slider, 0, 0, 1, 3)
        self.searchLay.addWidget(self.threshDisp, 1, 0, 1, 2)
        self.searchLay.addWidget(buttonSearch, 1, 2)
        buttonClearPts = QPushButton("Clear Points")
        buttonClearPts.clicked.connect(self._clearPts)

        self.cbAcquire = QCheckBox("Acquire")
        self.cmboxGroupPts = QComboBox()
        self.cmboxGroupPts.addItem("No Groups")
        self.cmboxGroupPts.addItem("Groups by radius")
        self.cmboxGroupPts.addItem("Entire mesh as one group")
        self.cmboxGroupPts.addItem("Number of Groups")
        self.cmboxGroupPts.addItem("Number of Points per Group")
        self.cmboxGroupPts.currentIndexChanged.connect(self._selectGroupOption)
        self.groupRadiusLabel = QLabel("Group Radius")
        self.groupRadiusLineEdit = QLineEdit()
        self.groupRadiusLineEdit.returnPressed.connect(
            lambda: self._setGroupRadius(self.groupRadiusLineEdit.text())
        )
        global navOptions
        if navOptions.groupRadius is not None:
            self._setGroupRadius(str(navOptions.groupRadius))
        self.groupRadiusLabelµm = QLabel("µm")
        self.pixelSizeLabel = QLabel("Pixel Size")
        self.pixelSizeLineEdit = QLineEdit()
        self.pixelSizeLineEdit.returnPressed.connect(
            lambda: self._setPixelSize(self.pixelSizeLineEdit.text())
        )
        if navOptions.pixelSize is not None:
            self._setPixelSize(str(navOptions.pixelSize))
        self.pixelSizeLabelnm = QLabel("nm")

        self.numGroupsLabel = QLabel("Number of groups")
        self.numGroupsSB = QSpinBox()
        self.numGroupsSB.setMinimum(1)
        self.numGroupsSB.setMaximum(20)
        self.numGroupsSB.setValue(navOptions.numGroups)
        self.numGroupsSB.valueChanged.connect(
            lambda: self._setNumGroups(self.numGroupsSB.value())
        )

        self.ptsPerGroupLabel = QLabel("Points per group")
        self.ptsPerGroupSB = QSpinBox()
        self.ptsPerGroupSB.setMinimum(1)
        self.ptsPerGroupSB.setMaximum(20)
        self.ptsPerGroupSB.setValue(navOptions.ptsPerGroup)
        self.ptsPerGroupSB.valueChanged.connect(
            lambda: self._setPtsPerGroup(self.ptsPerGroupSB.value())
        )

        buttonSaveAndQuit = QPushButton("\nSave and Quit\n")
        buttonSaveAndQuit.resize(buttonSaveAndQuit.sizeHint())
        buttonSaveAndQuit.clicked.connect(self.saveAndQuit)

        # layout
        vlay = QVBoxLayout()
        vlay.addWidget(self.crop_template)
        vlay.addWidget(self.cbBlurTemp)
        vlay.addWidget(self.cbBlurImg)
        vlay.addWidget(QLabel())
        vlay.addLayout(self.searchLay)
        vlay.addWidget(buttonClearPts)
        vlay.addWidget(QLabel())
        vlay.addWidget(self.cbAcquire)
        vlay.addWidget(QLabel("Grouping option"))
        vlay.addWidget(self.cmboxGroupPts)
        self.groupByRadiusLay = QGridLayout()
        self.groupByRadiusLay.addWidget(self.groupRadiusLabel, 1, 0)
        self.groupByRadiusLay.addWidget(self.groupRadiusLineEdit, 1, 1)
        self.groupByRadiusLay.addWidget(self.groupRadiusLabelµm, 1, 2)
        self.groupByRadiusLay.addWidget(self.pixelSizeLabel, 2, 0)
        self.groupByRadiusLay.addWidget(self.pixelSizeLineEdit, 2, 1)
        self.groupByRadiusLay.addWidget(self.pixelSizeLabelnm, 2, 2)
        vlay.addLayout(self.groupByRadiusLay)
        self.numGroupsLayout = QGridLayout()
        self.numGroupsLayout.addWidget(self.numGroupsLabel, 1, 0)
        self.numGroupsLayout.addWidget(self.numGroupsSB, 1, 2)
        vlay.addLayout(self.numGroupsLayout)
        self.ptsPerGroupLayout = QGridLayout()
        self.ptsPerGroupLayout.addWidget(self.ptsPerGroupLabel, 1, 0)
        self.ptsPerGroupLayout.addWidget(self.ptsPerGroupSB, 1, 2)
        vlay.addLayout(self.ptsPerGroupLayout)
        vlay.addWidget(QLabel())
        vlay.addWidget(buttonSaveAndQuit)

        self.cmboxGroupPts.setCurrentIndex(navOptions.groupOption)
        self._selectGroupOption(navOptions.groupOption)
        vlay.addStretch(1)
        self.setLayout(vlay)

    def _setThreshDisp(self, i: int):
        self.threshDisp.setValue(i / 10 ** self.sldPrec)

    def _setThreshSlider(self, val: float):
        try:
            self.slider.setValue(int(10 ** self.sldPrec * val))
            self.thresholdVal = val
        except ValueError:
            pass

    def _templateSearch(self):
        template = self.crop_template.originalImg
        image = self.parentWidget().viewer.originalImg
        if image.isNull() or template.isNull():
            popup(self, "either image or template missing")
            return

        self.parentWidget().parentWidget().statusBar().showMessage("searching...")
        startTime = time.time()
        global pts
        pts = templateMatch(
            qImgToNp(image),
            qImgToNp(template),
            self.thresholdVal,
            blurImage=self.cbBlurImg.isChecked(),
            blurTemplate=self.cbBlurTemp.isChecked(),
        )
        endTime = time.time()

        viewer = self.parentWidget().viewer
        viewer.searchedImg = drawCoords(viewer.originalImg, pts)
        viewer._setActiveImg(viewer.searchedImg)

        self.parentWidget().parentWidget().statusBar().showMessage(
            "%d matches found in %.2f sec" % (len(pts), endTime - startTime)
        )

        self.repaint()

    def _clearPts(self):
        global pts
        pts = []
        viewer = self.parentWidget().viewer
        viewer._setActiveImg(viewer.originalImg)
        viewer.searchedImg = QImage()
        viewer._refresh()
        self.parentWidget().parentWidget().statusBar().clearMessage()

    def saveAndQuit(self):
        global navOptions
        navOptions = NavOptions(
            navOptions.groupOption,
            navOptions.groupRadius,
            navOptions.pixelSize,
            navOptions.numGroups,
            navOptions.ptsPerGroup,
            int(self.cbAcquire.isChecked()),
        )
        global pts
        global finalPts
        finalPts = pts
        self.parentWidget().parentWidget().writeSettings()
        QApplication.quit()

    def _setGroupRadius(self, s: str):
        try:
            groupRadius = float("{:.1f}".format(float(s)))
            self.groupRadiusLineEdit.setText(str(groupRadius))
        except Exception as e:
            print("_setGroupRadius exception")
            print(e)
            groupRadius = None

        global navOptions
        navOptions = NavOptions(
            navOptions.groupOption,
            groupRadius,
            navOptions.pixelSize,
            navOptions.numGroups,
            navOptions.ptsPerGroup,
            navOptions.acquire,
        )

    def _setPixelSize(self, s: str):
        try:
            pixelSize = float("{:.1f}".format(float(s)))  # in nm
            self.pixelSizeLineEdit.setText(str(pixelSize))
        except Exception as e:
            print("_setPixelSize exception")
            print(e)
            pixelSize = None

        global navOptions
        navOptions = NavOptions(
            navOptions.groupOption,
            navOptions.groupRadius,
            pixelSize,
            navOptions.numGroups,
            navOptions.ptsPerGroup,
            navOptions.acquire,
        )

    def _setNumGroups(self, numGroups):
        global navOptions
        navOptions = NavOptions(
            navOptions.groupOption,
            navOptions.groupRadius,
            navOptions.pixelSize,
            numGroups,
            navOptions.ptsPerGroup,
            navOptions.acquire,
        )
        self.numGroupsSB.setValue(navOptions.numGroups)

    def _setPtsPerGroup(self, ptsPerGroup):
        global navOptions
        navOptions = NavOptions(
            navOptions.groupOption,
            navOptions.groupRadius,
            navOptions.pixelSize,
            navOptions.numGroups,
            ptsPerGroup,
            navOptions.acquire,
        )
        self.ptsPerGroupSB.setValue(navOptions.ptsPerGroup)

    def _selectGroupOption(self, groupOption):
        global navOptions
        navOptions = NavOptions(
            groupOption,
            navOptions.groupRadius,
            navOptions.pixelSize,
            navOptions.numGroups,
            navOptions.ptsPerGroup,
            navOptions.acquire,
        )
        if groupOption == 1:  # groups by radius
            self.groupRadiusLabel.show()
            self.groupRadiusLineEdit.show()
            self.groupRadiusLabelµm.show()
            self.pixelSizeLabel.show()
            self.pixelSizeLineEdit.show()
            self.pixelSizeLabelnm.show()
            self.numGroupsLabel.hide()
            self.numGroupsSB.hide()
            self.ptsPerGroupLabel.hide()
            self.ptsPerGroupSB.hide()
        elif groupOption == 3:  # numGroups
            self.groupRadiusLabel.hide()
            self.groupRadiusLineEdit.hide()
            self.groupRadiusLabelµm.hide()
            self.pixelSizeLabel.hide()
            self.pixelSizeLineEdit.hide()
            self.pixelSizeLabelnm.hide()
            self.numGroupsLabel.show()
            self.numGroupsSB.show()
            self.ptsPerGroupLabel.hide()
            self.ptsPerGroupSB.hide()
        elif groupOption == 4:  # ptsPerGroup
            self.groupRadiusLabel.hide()
            self.groupRadiusLineEdit.hide()
            self.groupRadiusLabelµm.hide()
            self.pixelSizeLabel.hide()
            self.pixelSizeLineEdit.hide()
            self.pixelSizeLabelnm.hide()
            self.numGroupsLabel.hide()
            self.numGroupsSB.hide()
            self.ptsPerGroupLabel.show()
            self.ptsPerGroupSB.show()
        else:
            self.groupRadiusLabel.hide()
            self.groupRadiusLineEdit.hide()
            self.groupRadiusLabelµm.hide()
            self.pixelSizeLabel.hide()
            self.pixelSizeLineEdit.hide()
            self.pixelSizeLabelnm.hide()
            self.numGroupsLabel.hide()
            self.numGroupsSB.hide()
            self.ptsPerGroupLabel.hide()
            self.ptsPerGroupSB.hide()
        self.repaint()

    def _refreshNavOptions(self):
        global navOptions
        self._setGroupRadius(navOptions.groupRadius)
        self._setPixelSize(navOptions.pixelSize)
        self._setNumGroups(navOptions.numGroups)
        self._setPtsPerGroup(navOptions.ptsPerGroup)
        self._selectGroupOption(navOptions.groupOption)
        self.cmboxGroupPts.setCurrentIndex(navOptions.groupOption)
        self.parentWidget().setAcquire(navOptions.acquire)


class MainWidget(QWidget):
    def __init__(self):
        super().__init__()
        self.initUI()

    def initUI(self):
        self.sidebar = Sidebar()
        self.viewer = ImageViewerCrop()

        grid = QGridLayout()
        grid.setSpacing(10)
        grid.addWidget(self.sidebar, 1, 0, 1, 1)
        grid.addWidget(self.viewer, 1, 1, 5, 5)
        self.setLayout(grid)

    def openImage(self, image):
        self.viewer.loadImage(image)

    def setTemplate(self, template):
        self.sidebar.crop_template.newImg(npToQImage(template))

    def setBlurImage(self, blur: bool):
        self.sidebar.cbBlurImg.setChecked(blur)

    def setBlurTemplate(self, blur: bool):
        self.sidebar.cbBlurTemp.setChecked(blur)

    def setAcquire(self, acquire: bool):
        self.sidebar.cbAcquire.setChecked(acquire)

    def setThreshold(self, theshold):
        self.sidebar.threshDisp.setValue(theshold)

    def search(self):
        self.sidebar._templateSearch()


class MainWindow(QMainWindow):
    def __init__(self):
        super().__init__()
        self.root = MainWidget()
        self.setCentralWidget(self.root)
        self.statusBar()
        self.initUI()
        self.readSettings()

    def initUI(self):
        menubar = self.menuBar()
        fileMenu = menubar.addMenu("File")
        viewMenu = menubar.addMenu("View")

        zoomIn = QAction("Zoom In", self)
        zoomIn.setShortcut(Qt.Key_Equal)
        zoomIn.triggered.connect(self.root.viewer.zoomIn)
        zoomOut = QAction("Zoom Out", self)
        zoomOut.setShortcut(Qt.Key_Minus)
        zoomOut.triggered.connect(self.root.viewer.zoomOut)
        viewMenu.addAction(zoomIn)
        viewMenu.addAction(zoomOut)

        self.showMaximized()

    def openImage(self, image):
        self.root.openImage(image)

    def setTemplate(self, template):
        self.root.setTemplate(template)

    def setBlurImage(self, blur: bool):
        self.root.setBlurImage(blur)

    def setBlurTemplate(self, blur: bool):
        self.root.setBlurTemplate(blur)

    def setAcquire(self, acquire: bool):
        self.root.setAcquire(acquire)

    def getThreshold(self):
        return self.root.sidebar.thresholdVal

    def setThreshold(self, theshold):
        self.root.setThreshold(theshold)

    def setGroupOption(self, option):
        self.root.sidebar._selectGroupOption(option)
        self.root.sidebar.cmboxGroupPts.setCurrentIndex(option)

    def search(self):
        self.root.search()

    def writeSettings(self):
        settings = QSettings(
            "University of Massachusetts Medical School CryoEM", "semmatch"
        )
        settings.beginGroup("MainWindow")
        settings.setValue("size", self.size())
        settings.setValue("pos", self.pos())
        settings.setValue("threshold", self.getThreshold())
        settings.setValue("blurImage", self.root.sidebar.cbBlurImg.isChecked())
        settings.setValue("blurTemplate", self.root.sidebar.cbBlurTemp.isChecked())
        global navOptions
        settings.setValue("navOptions", navOptions)
        settings.endGroup()

    def readSettings(self):
        settings = QSettings(
            "University of Massachusetts Medical School CryoEM", "semmatch"
        )
        settings.beginGroup("MainWindow")
        self.resize(settings.value("size", QSize(400, 400)))
        self.move(settings.value("pos", QPoint(200, 200)))
        global inputThreshold
        self.setThreshold(float(settings.value("threshold", inputThreshold)))
        global navOptions
        navOptions = settings.value("navOptions", navOptions)
        self.root.sidebar._refreshNavOptions()
        self.setBlurImage(settings.value("blurImage") == "true")
        self.setBlurTemplate(settings.value("blurTemplate") == "true")
        settings.endGroup()


def main(
    image, template, threshold, options: "NavOptions", blurImage=True, blurTemplate=True
):
    global inputThreshold
    inputThreshold = threshold

    global navOptions
    navOptions = options

    global pts
    pts = []

    global finalPts
    finalPts = []

    QApplication.setAttribute(Qt.AA_EnableHighDpiScaling)
    app = QApplication([])
    w = MainWindow()
    w.openImage(image)
    if template is not None:
        w.setTemplate(template)
        w.search()

    app.exec_()

    return finalPts, navOptions



import io
import cv2
import numpy as np
import PIL
import PIL.ImageQt
from PyQt5.QtGui import QPixmap
from PyQt5.QtCore import QBuffer

# Unset PIL max size
PIL.Image.MAX_IMAGE_PIXELS = None


def npToQImage(ndArr):
    pilImageQt = PIL.ImageQt.ImageQt(PIL.Image.fromarray(ndArr))
    return QPixmap.fromImage(pilImageQt).toImage()


def qImgToPilRGBA(qimg):
    buf = QBuffer()
    buf.open(QBuffer.ReadWrite)
    qimg.save(buf, "PNG")
    return PIL.Image.open(io.BytesIO(buf.data().data())).convert("RGBA")


def qImgToNp(qimg):
    return np.array(qImgToPilRGBA(qimg))


def drawCross(img: "ndarray", x, y):
    red = (255, 0, 0, 255)
    thickness = 2
    cv2.line(img, (x - 15, y), (x + 15, y), red, thickness)
    cv2.line(img, (x, y - 15), (x, y + 15), red, thickness)


def drawCrosses(img: "ndarray", coords):
    img = np.flip(img, 0).copy()
    for x, y in coords:
        drawCross(img, x, y)
    return np.flip(img, 0).copy()


def drawCoords(qimg, coords):
    return npToQImage(drawCrosses(qImgToNp(qimg), coords))



from collections import namedtuple
import math

import cv2
import numpy as np
import PIL
from PIL import Image, ImageFilter
from scipy.ndimage.filters import gaussian_filter, maximum_filter
import skimage

# ubiquitous Point type
Pt = namedtuple("Pt", "x y")

# output nav grouping/acquire options
NavOptions = namedtuple(
    "NavOptions", "groupOption groupRadius pixelSize numGroups ptsPerGroup acquire"
)


# for relative distance, square distance is faster to compute
def squareDist(pt1, pt2):
    return (pt1.x - pt2.x) ** 2 + (pt1.y - pt2.y) ** 2


# prevent including the same hole multiple times
def pointsExistWithinRadius(center: "Pt", coords, radius):
    radius_2 = radius ** 2
    if len(coords) == 0:
        return False
    for pt in coords:
        if squareDist(pt, center) < radius_2:
            return True
    return False


# modified from OpenCV docs
# https://docs.opencv.org/3.4/d4/dc6/tutorial_py_template_matching.html
def templateMatch(
    image,
    template,
    threshold,
    downSample: int = 1,
    blurImage=False,
    blurTemplate=False,
    sigma=10,
):
    """Return a list of (x,y) pixel coordinates where cross-correlation between
    the image and template surpass the threshold value.

    To prevent double counting, the minimum distance between coordinates is
    constrained to the length of the larger dimension of the template image.

    Following SerialEM conventions, (0,0) is at the bottom-left corner,
    with +x axis to the right and +y axis upwards.

    Images can be downsampled for faster computation and noise reduction.
    """

    if len(image.shape) == 3:
        image = image[:, :, 0]
    if len(template.shape) == 3:
        template = template[:, :, 0]
    image = image[::downSample, ::downSample]
    template = template[::downSample, ::downSample]
    if blurImage:
        image = gaussian_filter(image, sigma=sigma)
    if blurTemplate:
        template = gaussian_filter(template, sigma=sigma)

    # flip both arrays upsidedown for coordinate conventions
    image = np.flip(image, 0).copy()
    template = np.flip(template, 0).copy()
    h, w, *_ = template.shape
    xcorrScores = cv2.matchTemplate(image, template, cv2.TM_CCOEFF_NORMED)
    maxfilter = maximum_filter(
        xcorrScores, size=(template.shape[0] // 2, template.shape[1] // 2)
    )
    loc = zip(*np.where((xcorrScores >= threshold) & (xcorrScores == maxfilter)))

    scoresIndex = [(x, y, xcorrScores[y][x]) for y, x in loc]
    scoresIndex.sort(key=lambda a: a[2], reverse=True)

    matches = []
    for x, y, _ in scoresIndex:
        x += w // 2
        y += h // 2
        if not pointsExistWithinRadius(Pt(x, y), matches, radius=max(h, w)):
            matches.append(Pt(x, y))
    # multiply back to get correct coordinates
    matches = [Pt(downSample * x, downSample * y) for x, y in matches]
    return matches


def imresize(img: "ndarray", factor):
    return np.array(
        PIL.Image.fromarray(img).resize(
            size=(int(img.shape[1] * factor), int(img.shape[0] * factor)),
            resample=PIL.Image.LANCZOS,
        )
    )


# https://pastebin.com/sBsPX4Y7
def anisodiff(
    img, niter=10, kappa=50, gamma=0.1, step=(1.0, 1.0), option=1, ploton=False
):
    """
        Anisotropic diffusion.

        Usage:
        imgout = anisodiff(im, niter, kappa, gamma, option)

        Arguments:
                img    - input image
                niter  - number of iterations
                kappa  - conduction coefficient 20-100 ?
                gamma  - max value of .25 for stability
                step   - tuple, the distance between adjacent pixels in (y,x)
                option - 1 Perona Malik diffusion equation No 1
                         2 Perona Malik diffusion equation No 2
                ploton - if True, the image will be plotted on every iteration

        Returns:
                imgout   - diffused image.

        kappa controls conduction as a function of gradient.  If kappa is low
        small intensity gradients are able to block conduction and hence diffusion
        across step edges.  A large value reduces the influence of intensity
        gradients on conduction.

        gamma controls speed of diffusion (you usually want it at a maximum of
        0.25)

        step is used to scale the gradients in case the spacing between adjacent
        pixels differs in the x and y axes

        Diffusion equation 1 favours high contrast edges over low contrast ones.
        Diffusion equation 2 favours wide regions over smaller ones.

        Reference:
        P. Perona and J. Malik.
        Scale-space and edge detection using ansotropic diffusion.
        IEEE Transactions on Pattern Analysis and Machine Intelligence,
        12(7):629-639, July 1990.

        Original MATLAB code by Peter Kovesi
        School of Computer Science & Software Engineering
        The University of Western Australia
        pk @ csse uwa edu au
        <http://www.csse.uwa.edu.au>

        Translated to Python and optimised by Alistair Muldal
        Department of Pharmacology
        University of Oxford
        <alistair.muldal@pharm.ox.ac.uk>

        June 2000  original version.
        March 2002 corrected diffusion eqn No 2.
        July 2012 translated to Python
        """

    # ...you could always diffuse each color channel independently if you
    # really want
    if img.ndim == 3:
        warnings.warn("Only grayscale images allowed, converting to 2D matrix")
        img = img.mean(2)

    # initialize output array
    img = img.astype("float32")
    imgout = img.copy()

    # initialize some internal variables
    deltaS = np.zeros_like(imgout)
    deltaE = deltaS.copy()
    NS = deltaS.copy()
    EW = deltaS.copy()
    gS = np.ones_like(imgout)
    gE = gS.copy()

    # create the plot figure, if requested
    if ploton:
        import pylab as pl
        from time import sleep

        fig = pl.figure(figsize=(20, 5.5), num="Anisotropic diffusion")
        ax1, ax2 = fig.add_subplot(1, 2, 1), fig.add_subplot(1, 2, 2)

        ax1.imshow(img, interpolation="nearest")
        ih = ax2.imshow(imgout, interpolation="nearest", animated=True)
        ax1.set_title("Original image")
        ax2.set_title("Iteration 0")

        fig.canvas.draw()

    for ii in range(niter):

        # calculate the diffs
        deltaS[:-1, :] = np.diff(imgout, axis=0)
        deltaE[:, :-1] = np.diff(imgout, axis=1)

        # conduction gradients (only need to compute one per dim!)
        if option == 1:
            gS = np.exp(-(deltaS / kappa) ** 2.0) / step[0]
            gE = np.exp(-(deltaE / kappa) ** 2.0) / step[1]
        elif option == 2:
            gS = 1.0 / (1.0 + (deltaS / kappa) ** 2.0) / step[0]
            gE = 1.0 / (1.0 + (deltaE / kappa) ** 2.0) / step[1]

        # update matrices
        E = gE * deltaE
        S = gS * deltaS

        # subtract a copy that has been shifted 'North/West' by one
        # pixel. don't as questions. just do it. trust me.
        NS[:] = S
        EW[:] = E
        NS[1:, :] -= S[:-1, :]
        EW[:, 1:] -= E[:, :-1]

        # update the image
        imgout += gamma * (NS + EW)

        if ploton:
            iterstring = "Iteration %i" % (ii + 1)
            ih.set_data(imgout)
            ax2.set_title(iterstring)
            fig.canvas.draw()
            # sleep(0.01)

    return imgout.astype(np.uint8)


def median_filt(img, radius=5):
    return np.array(Image.fromarray(img).filter(ImageFilter.MedianFilter(size=radius)))


def scharr(img):
    return (2000 * skimage.filters.scharr(img)).astype("uint8")


def prefilter_before_hough(img):
    img = anisodiff(img, niter=20)
    img = median_filt(img)
    img = scharr(img)
    return img


def houghCircles(img, param1=50, param2=60, minDist=30, minRadius=5, maxRadius=45):
    img = prefilter_before_hough(img)

    circles = cv2.HoughCircles(
        img,
        cv2.HOUGH_GRADIENT,
        dp=1,
        minDist=minDist,
        param1=param1,
        param2=param2,
        minRadius=minRadius,
        maxRadius=maxRadius,
    )
    try:
        circles = np.uint16(np.around(circles))
    except AttributeError:
        circles = None

    if circles is not None:
        pts = [(x, img.shape[0] - y) for x, y in circles[0][:, :2]]
    else:
        pts = []

    return pts



import numpy as np
import PIL
from scipy.ndimage import gaussian_filter
from semmatch.core import templateMatch, Pt, NavOptions
from semmatch.autodoc import ptsToNavPts, openNavfile


def test_templateMatch():
    MMM = np.array(PIL.Image.open("MMM.jpg"))
    MMM = gaussian_filter(MMM, sigma=1)

    template = np.array(PIL.Image.open("T.jpg"))
    template = gaussian_filter(template, sigma=1)

    coords = templateMatch(MMM, template, threshold=0.8)
    assert coords == [
        (2149, 1904),
        (2637, 1448),
        (1732, 1229),
        (2239, 1535),
        (2293, 2128),
        (2897, 1120),
        (2111, 2088),
        (1769, 1044),
        (2819, 1488),
        (2019, 1680),
        (2095, 1311),
        (2598, 1633),
    ]


def test_writeToNavFile():
    coords = [
        (2149, 1904),
        (2637, 1448),
        (1732, 1229),
        (2239, 1535),
        (2293, 2128),
        (2897, 1120),
        (2111, 2088),
        (1769, 1044),
        (2819, 1488),
        (2019, 1680),
        (2095, 1311),
        (2598, 1633),
    ]
    coords = [Pt(*coord) for coord in coords]
    mapID = "30-A"
    nav = openNavfile("nav.nav")
    options = NavOptions(groupOption=0, groupRadius=123, pixelSize=1, numGroups=1, acquire=1)
    navPts = ptsToNavPts(coords, nav, mapID, startLabel=9000, options=options)
    newNavData = "AdocVersion = 2.00\n\n" + "".join(str(pt) for pt in navPts)
    with open("newNav.nav", "w") as f:
        f.write(newNavData)

    with open("newNav_validate.nav") as f:
        validationData = f.read()

    assert validationData == newNavData



import functools
import math
import operator
import random

import numpy as np
from sklearn.cluster import KMeans
from semmatch.core import squareDist, Pt


# https://stackoverflow.com/a/51075698
# clockwise sort starting from bottom left
def clockwiseSort(pts):
    center = tuple(
        map(
            operator.truediv,
            functools.reduce(lambda x, y: map(operator.add, x, y), pts),
            [len(pts)] * 2,
        )
    )
    return sorted(
        pts,
        key=lambda pt: (
            -135 - math.degrees(math.atan2(*tuple(map(operator.sub, pt, center))[::-1]))
        )
        % 360,
    )


def centroid(pts: "ndarray"):
    length = pts.shape[0]
    sum_x = np.sum(pts[:, 0])
    sum_y = np.sum(pts[:, 1])
    return Pt(sum_x / length, sum_y / length)


def closestPtToCentroid(pts):
    """Returns the coordinate closest to the center of mass"""
    center = centroid(np.array(pts))
    closestPoint = pts[0]
    msd = squareDist(pts[0], center)  # min square distance
    for pt in pts:
        dist_2 = squareDist(pt, center)
        if dist_2 < msd:
            closestPoint = pt
            msd = dist_2
    return closestPoint


def greedyPathThroughPts(coords):
    """Returns a list with the first item being the left most coordinate,
       and successive items being the minimum distance from the previous item.
    """
    if coords == []:
        return []

    coords = [Pt(*coord) for coord in coords]
    leftMostPt = sorted(coords, key=lambda x: x[0])[0]
    unvisitedPts = set(coords)
    unvisitedPts.remove(leftMostPt)

    result = [leftMostPt]
    while unvisitedPts:
        closestPtToPrev = unvisitedPts.pop()
        unvisitedPts.add(closestPtToPrev)
        minDist = squareDist(closestPtToPrev, result[-1])
        for pt in unvisitedPts:
            distFromPrev = squareDist(pt, result[-1])
            if distFromPrev < minDist:
                minDist = distFromPrev
                closestPtToPrev = pt
        result.append(closestPtToPrev)
        unvisitedPts.remove(closestPtToPrev)
    return result


def makeGroupsOfPoints(pts, max_radius):
    max_radius_2 = max_radius ** 2
    groups = []
    greedyPath = greedyPathThroughPts(pts)[::-1]
    try:
        group = [greedyPath.pop()]
    except:
        print("pts list is empty")
        return []
    while greedyPath:
        pt = greedyPath.pop()
        if squareDist(pt, group[0]) < max_radius_2:
            group.append(pt)
        else:
            groups.append(group)
            group = [pt]
    if group:
        groups.append(group)

    for i in range(len(groups)):
        groupLeader = closestPtToCentroid(groups[i])
        groups[i] = [groupLeader] + [pt for pt in groups[i] if pt != groupLeader]
    return groups


def k_means(pts, k):
    if len(pts) < k:
        print(
            "only %d points were found, which is less than %d (numGroups); setting number of groups to 1"
            % (len(pts), k)
        )
        k = 1
        if len(pts) == 1:
            return [pts]
    labels = KMeans(k).fit(pts).labels_
    groups = []
    for i in range(k):
        group = []
        for j in np.where(labels == i)[0]:
            group.append(pts[j])
        if len(group) > 0:
            groups.append(group)
    return groups

def getRandPts(pts, maxPts):
    try:
        return random.sample(pts, maxPts)
    except ValueError:
        return pts



import random
import os
from semmatch.groups import (
    greedyPathThroughPts,
    makeGroupsOfPoints,
    k_means,
    closestPtToCentroid,
)


def openNavfile(navfile) -> dict:
    nav = {}
    try:
        with open(navfile) as f:
            data = f.read()
    except FileNotFoundError:
        print(
            "could not find %s in current working directory: %s"
            % (navfile, os.getcwd())
        )
        exit()
    sections = data.split("\n\n")
    if "AdocVersion" not in sections[0]:
        raise Exception("could not find AdocVersion")
    sections = [section.split("\n") for section in sections[1:]]
    sections = list(filter(lambda x: x != [""], sections))
    for section in sections:
        item = section[0][1:-1].split("=")[1].strip()
        sectionData = {}
        for line in section[1:]:
            if "=" in line:
                key, val = line.split("=", 1)
                key = key.strip()
                val = val.strip()
                sectionData[key] = val
        nav[item] = sectionData
    return nav


def isValidAutodoc(navfile):
    try:
        with open(navfile) as f:
            for line in f:
                if line.strip():
                    if line.split()[0] == "AdocVersion":
                        return True
                    else:
                        print("error: could not find AdocVersion")
                        return False
    except:
        print("invalid autodoc")
        return False


def isValidLabel(data: "list", label: str):
    try:
        mapSectionIndex = data.index(f"[Item = {label}]")
    except:
        print("unable to write new autodoc file: label %s not found" % label)
        return False
    return True


class NavFilePoint:
    def __init__(
        self,
        label: str,
        regis: int,
        ptsX: int,
        ptsY: int,
        zHeight: float,
        drawnID: int,
        numPts: int = 1,
        itemType: int = 0,
        color: int = 0,
        groupID: int = 0,
        acquire: int = 0,
        **kwargs,
    ):
        self._label = label
        self.Color = color
        self.NumPts = numPts
        self.Regis = regis
        self.Type = itemType
        self.PtsX = ptsX
        self.PtsY = ptsY
        self.DrawnID = drawnID
        self.GroupID = groupID
        self.Acquire = acquire
        self.CoordsInMap = [ptsX, ptsY, zHeight]
        vars(self).update(kwargs)

    def __str__(self):
        result = [f"[Item = {self._label}]"]
        for key, val in vars(self).items():
            if key == "_label":
                continue
            if key == "CoordsInMap":
                val = " ".join(str(x) for x in val)
            result.append(f"{key} = {val}")
        result.append("\n")
        return "\n".join(result)


def ptsToNavPts(
    coords, nav: dict, mapLabel: str, startLabel: int, options: "NavOptions"
):
    try:
        regis = int(nav[mapLabel]["Regis"])
        drawnID = int(nav[mapLabel]["MapID"])
        zHeight = float(nav[mapLabel]["StageXYZ"].split()[2])
    except KeyError as e:
        print(e)
        exit()

    navPoints = []
    label = startLabel

    if options.groupOption == 0:  # no groups
        for pt in greedyPathThroughPts(coords):
            navPoints.append(
                NavFilePoint(
                    label, regis, *pt, zHeight, drawnID, acquire=options.acquire
                )
            )
            label += 1
    elif options.groupOption == 1:  # groups withing mesh
        try:
            groupRadiusPix = options.groupRadius * 1000 / options.pixelSize
        except ZeroDivisionError:
            print("pixel size can't be 0; aborting")
            exit()

        for group in makeGroupsOfPoints(coords, groupRadiusPix):
            group = greedyPathThroughPts(group)
            subLabel = 1
            groupID = random.randint(10 ** 9, 2 * 10 ** 9)
            for pt in group:
                navPoints.append(
                    NavFilePoint(
                        f"{label}-{subLabel}",
                        regis,
                        *pt,
                        zHeight,
                        drawnID,
                        groupID=groupID,
                        acquire=options.acquire,
                    )
                )
                subLabel += 1
            label += 1
    elif options.groupOption == 2:  # entire mesh as group
        groupID = random.randint(10 ** 9, 2 * 10 ** 9)
        subLabel = 1
        for pt in greedyPathThroughPts(coords):
            navPoints.append(
                NavFilePoint(
                    f"{label}-{subLabel}",
                    regis,
                    *pt,
                    zHeight,
                    drawnID,
                    groupID=groupID,
                    acquire=options.acquire,
                )
            )
            subLabel += 1
        label += 1
    elif options.groupOption == 3:  # numGroups
        groups = []
        for group in k_means(coords, options.numGroups):
            groupLeader = closestPtToCentroid(group)
            group.remove(groupLeader)
            group = [groupLeader] + greedyPathThroughPts(group)
            groups.append(group)
        groups.sort(key=lambda group: group[0][0])  # sort by group leader's x position

        for group in groups:
            subLabel = 1
            groupID = random.randint(10 ** 9, 2 * 10 ** 9)
            for pt in group:
                navPoints.append(
                    NavFilePoint(
                        f"{label}-{subLabel}",
                        regis,
                        *pt,
                        zHeight,
                        drawnID,
                        groupID=groupID,
                        acquire=options.acquire,
                    )
                )
                subLabel += 1
            label += 1
    elif options.groupOption == 4:  # points per group
        groups = []
        numGroups = len(coords) // options.ptsPerGroup
        if numGroups == 0:
            numGroups = 1
        for group in k_means(coords, numGroups):
            groupLeader = closestPtToCentroid(group)
            group.remove(groupLeader)
            group = [groupLeader] + greedyPathThroughPts(group)
            groups.append(group)
        groups.sort(key=lambda group: group[0][0])  # sort by group leader's x position

        for group in groups:
            subLabel = 1
            groupID = random.randint(10 ** 9, 2 * 10 ** 9)
            for pt in group:
                navPoints.append(
                    NavFilePoint(
                        f"{label}-{subLabel}",
                        regis,
                        *pt,
                        zHeight,
                        drawnID,
                        groupID=groupID,
                        acquire=options.acquire,
                    )
                )
                subLabel += 1
            label += 1
    else:
        raise ValueError("groupOption needs to be 0, 1, 2, 3 or 4")
    return navPoints


def createAutodoc(outputfile, navPts):
    with open(outputfile, "w") as f:
        f.write("AdocVersion = 2.00\n\n" + "".join(str(pt) for pt in navPts))



def main():
    import argparse
    import sys

    import imageio

    from semmatch.core import Pt, NavOptions, templateMatch, imresize, houghCircles
    from semmatch.autodoc import ptsToNavPts, createAutodoc, openNavfile
    from semmatch.groups import getRandPts

    parser = argparse.ArgumentParser(description="template matching tool for SerialEM")
    # required
    parser.add_argument("--navfile", help="SerialEM nav file", required=True)
    parser.add_argument("--image", help="jpg", required=True)
    parser.add_argument("--mapLabel", help="label id", required=True)
    parser.add_argument(
        "--newLabel", help="starting label of added points", type=int, required=True
    )
    parser.add_argument("-o", "--output", help="output nav file", required=True)

    # optional
    parser.add_argument("--gui", help="interactive gui mode", action="store_true")
    parser.add_argument(
        "--template", help="template image to use; required in non-gui mode"
    )
    parser.add_argument(
        "--houghCircles", help="automatic detection", action="store_true"
    )
    parser.add_argument(
        "--param2", help="threshold for houghCircles", type=int, default=60
    )
    parser.add_argument(
        "--maxPts", help="limit number of pts found via houghCircles", type=int
    )
    parser.add_argument(
        "--groupOption",
        help="grouping option for points;"
        + "0 = no groups; "
        + "1 = groups based on radius (requires --groupRadius and --pixelSize);"
        + "2 = all points as one group;"
        + "3 = specify number of groups;"
        + "4 = specify number of points per group",
        type=int,
        default=0,
    )
    parser.add_argument(
        "--threshold", help="threshold value for zncc", type=float, default=0.8
    )
    parser.add_argument(
        "--reduction", help="external reduction factor", type=float, default=1.0
    )
    parser.add_argument(
        "--acquire", help="mark points with acquire flag", type=int, default=1
    )
    parser.add_argument(
        "--groupRadius", help="groupRadius in µm", type=float, default=7.0
    )
    parser.add_argument("--pixelSize", help="pixelSize in nm", type=float, default=5.0)
    parser.add_argument(
        "--numGroups",
        help="number of groups for k-means groupOption",
        type=int,
        default=10,
    )
    parser.add_argument(
        "--ptsPerGroup", help="specify number of points per group", type=int, default=8
    )
    parser.add_argument("--noBlurImage", help="", action="store_true")
    parser.add_argument("--noBlurTemplate", help="", action="store_true")

    args = parser.parse_args()

    navfile = args.navfile
    image = args.image
    reduction = args.reduction
    mapLabel = args.mapLabel
    newLabel = args.newLabel
    output = args.output
    template = args.template
    threshold = args.threshold
    groupOption = args.groupOption
    groupRadius = args.groupRadius
    pixelSize = args.pixelSize
    numGroups = args.numGroups
    ptsPerGroup = args.ptsPerGroup
    acquire = args.acquire
    blurImage = not args.noBlurImage
    blurTemplate = not args.noBlurTemplate
    options = NavOptions(
        groupOption, groupRadius, pixelSize, numGroups, ptsPerGroup, acquire
    )
    param2 = args.param2
    maxPts = args.maxPts

    # clear output file to prevent merging previous points
    createAutodoc(output, [])

    # unnecessary options messages
    if groupOption != 1:
        if groupRadius is not None:
            print(
                "groupRadius will be ignored because groupOption is not 1 (groups by radius)"
            )
        if pixelSize is not None:
            print(
                "pixelSize will be ignored because groupOption is not 1 (groups by radius)"
            )
    if groupOption != 3 and numGroups is not None:
        print(
            "numGroups will be ignored because groupOption is not 3 (specify number of groups)"
        )
    if groupOption != 4 and ptsPerGroup is not None:
        print(
            "ptsPerGroup will be ignored because groupOption is not 4 (specify number of points per group)"
        )

    nav = openNavfile(navfile)
    if mapLabel not in nav:
        print("could not find map label: %s; aborting" % mapLabel)
        exit()
    elif not (
        "Regis" in nav[mapLabel]
        and "MapID" in nav[mapLabel]
        and "StageXYZ" in nav[mapLabel]
    ):
        print(
            "either Regis, MapID, and/or StageXYZ missing in section labeled %s; aborting"
            % mapLabel
        )
        exit()

    # read and downsize images if necessary
    image = imageio.imread(image)
    if template is not None:
        try:
            template = imageio.imread(template)
            template = imresize(
                template, 1 / (reduction)
            )
        except Exception as e:
            print(e)
            print(
                "error reading in template %s; continuing without template" % template
            )
            template = None

    if args.houghCircles == True:
        print("using hough circles")
        pts = houghCircles(image, param2=param2)
        if maxPts is not None:
            pts = getRandPts(pts, maxPts)
    elif args.gui == True:
        print("using template matching gui")
        import semmatch.gui

        pts, options = semmatch.gui.main(
            image,
            template,
            threshold,
            options,
            blurImage=blurImage,
            blurTemplate=blurTemplate,
        )
        if options.groupRadius is None:
            print("invalid group radius; aborting")
            exit()
        if options.pixelSize is None:
            print("invalid pixel size; aborting")
            exit()
    else:
        print("using template matching")
        if template is None:
            print("non-gui option must specify template")
            exit()
        pts = templateMatch(
            image, template, threshold, blurImage=blurImage, blurTemplate=blurTemplate
        )

    # compensate round off error from reduction
    pts = [Pt(x + 2, y) for x, y in pts]
    pts = [Pt(int(reduction * x), int(reduction * y)) for x, y in pts]

    if len(pts) == 0:
        print("no matches found; exiting without creating %s" % output)
        exit()
    navPts = ptsToNavPts(pts, nav, mapLabel, newLabel, options)
    createAutodoc(output, navPts)
    print("%s created" % output)


if __name__ == "__main__":
    main()



import setuptools

setuptools.setup(
    name="semmatch",
    version="0.0.12",
    author="Albert Xu",
    author_email="albert.t.xu@gmail.com",
    description="template matching tool for SerialEM",
    packages=setuptools.find_packages(),
    install_requires=[
        "imageio",
        "numpy",
        "opencv-python",
        "PyQt5",
        "Pillow",
        "scikit-image",
        "scikit-learn",
        "scipy",
    ],
    entry_points={"console_scripts": ["semmatch = semmatch.__main__:main"]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
